"use strict";Object.defineProperty(exports, "__esModule", {value: true});/**
 * Modlog
 * Pokemon Showdown - http://pokemonshowdown.com/
 *
 * Moderator actions are logged into a set of files known as the moderation log, or "modlog."
 * This file handles reading, writing, and querying the modlog.
 *
 * @license MIT
 */

var _child_process = require('child_process'); var child_process = _child_process;
var _path = require('path');
var _util = require('util'); var util = _util;

var _fs = require('../.lib-dist/fs');
var _processmanager = require('../.lib-dist/process-manager');
var _repl = require('../.lib-dist/repl');

var _converter = require('../tools/modlog/converter');

const MAX_PROCESSES = 1;
// If a modlog query takes longer than this, it will be logged.
const LONG_QUERY_DURATION = 2000;
const MODLOG_PATH = 'logs/modlog';

const GLOBAL_PUNISHMENTS = [
	'WEEKLOCK', 'LOCK', 'BAN', 'RANGEBAN', 'RANGELOCK', 'FORCERENAME',
	'TICKETBAN', 'AUTOLOCK', 'AUTONAMELOCK', 'NAMELOCK', 'AUTOBAN', 'MONTHLOCK',
];
const GLOBAL_PUNISHMENTS_REGEX_STRING = `\\b(${GLOBAL_PUNISHMENTS.join('|')}):.*`;

const PUNISHMENTS = [
	...GLOBAL_PUNISHMENTS, 'ROOMBAN', 'WEEKROOMBAN', 'UNROOMBAN', 'WARN', 'MUTE', 'HOURMUTE', 'UNMUTE',
	'CRISISDEMOTE', 'UNLOCK', 'UNLOCKNAME', 'UNLOCKRANGE', 'UNLOCKIP', 'UNBAN',
	'UNRANGEBAN', 'TRUSTUSER', 'UNTRUSTUSER', 'BLACKLIST', 'BATTLEBAN', 'UNBATTLEBAN',
	'NAMEBLACKLIST', 'KICKBATTLE', 'UNTICKETBAN', 'HIDETEXT', 'HIDEALTSTEXT', 'REDIRECT',
	'NOTE', 'MAFIAHOSTBAN', 'MAFIAUNHOSTBAN', 'GIVEAWAYBAN', 'GIVEAWAYUNBAN',
	'TOUR BAN', 'TOUR UNBAN', 'UNNAMELOCK',
];
const PUNISHMENTS_REGEX_STRING = `\\b(${PUNISHMENTS.join('|')}):.*`;

const execFile = util.promisify(child_process.execFile);







































class SortedLimitedLengthList {
	
	

	constructor(maxSize) {
		this.maxSize = maxSize;
		this.list = [];
	}

	getListClone() {
		return this.list.slice();
	}

	insert(element) {
		let insertedAt = -1;
		for (let i = this.list.length - 1; i >= 0; i--) {
			if (element.localeCompare(this.list[i]) < 0) {
				insertedAt = i + 1;
				if (i === this.list.length - 1) {
					this.list.push(element);
					break;
				}
				this.list.splice(i + 1, 0, element);
				break;
			}
		}
		if (insertedAt < 0) this.list.splice(0, 0, element);
		if (this.list.length > this.maxSize) {
			this.list.pop();
		}
	}
}

 function checkRipgrepAvailability() {
	if (Config.ripgrepmodlog === undefined) {
		Config.ripgrepmodlog = (async () => {
			try {
				await execFile('rg', ['--version'], {cwd: _path.normalize.call(void 0, `${__dirname}/../`)});
				await execFile('tac', ['--version'], {cwd: _path.normalize.call(void 0, `${__dirname}/../`)});
				return true;
			} catch (error) {
				return false;
			}
		})();
	}
	return Config.ripgrepmodlog;
} exports.checkRipgrepAvailability = checkRipgrepAvailability;

 class Modlog {
	
	/**
	 * If a stream is undefined, that means it has not yet been initialized.
	 * If a stream is truthy, it is open and ready to be written to.
	 * If a stream is null, it has been destroyed/disabled.
	 */
	__init() {this.sharedStreams = new Map()}
	__init2() {this.streams = new Map()}

	constructor(path) {;Modlog.prototype.__init.call(this);Modlog.prototype.__init2.call(this);
		this.logPath = path;
	}

	/**************************************
	 * Methods for writing to the modlog. *
	 **************************************/
	initialize(roomid) {
		if (this.streams.get(roomid)) return;
		const sharedStreamId = this.getSharedID(roomid);
		if (!sharedStreamId) {
			return this.streams.set(roomid, _fs.FS.call(void 0, `${this.logPath}/modlog_${roomid}.txt`).createAppendStream());
		}

		let stream = this.sharedStreams.get(sharedStreamId);
		if (!stream) {
			stream = _fs.FS.call(void 0, `${this.logPath}/modlog_${sharedStreamId}.txt`).createAppendStream();
			this.sharedStreams.set(sharedStreamId, stream);
		}
		this.streams.set(roomid, stream);
	}

	getSharedID(roomid) {
		return roomid.includes('-') ? `${toID(roomid.split('-')[0])}-rooms`  : false;
	}

	/**
	 * Writes to the modlog
	 */
	write(roomid, entry, overrideID) {
		roomid = entry.roomID || roomid;
		const stream = this.streams.get(roomid );
		if (!stream) throw new Error(`Attempted to write to an uninitialized modlog stream for the room '${roomid}'`);

		let buf = `[${new Date(entry.time || Date.now()).toJSON()}] (${overrideID || entry.visualRoomID || roomid}) ${entry.action}:`;
		if (entry.userid) buf += ` [${entry.userid}]`;
		if (entry.autoconfirmedID) buf += ` ac:[${entry.autoconfirmedID}]`;
		if (entry.alts) buf += ` alts:[${entry.alts.join('], [')}]`;
		if (entry.ip) buf += ` [${entry.ip}]`;
		if (entry.loggedBy) buf += ` by ${entry.loggedBy}`;
		if (entry.note) buf += `: ${entry.note}`;

		void stream.write(`${buf}\n`);
	}

	async destroy(roomid) {
		const stream = this.streams.get(roomid);
		if (stream && !this.getSharedID(roomid)) {
			this.streams.set(roomid, null);
			await stream.writeEnd();
		}
		this.streams.set(roomid, null);
	}

	async destroyAll() {
		const promises = [];
		for (const id in this.streams) {
			promises.push(this.destroy(id ));
		}
		return Promise.all(promises);
	}

	async rename(oldID, newID) {
		const streamExists = this.streams.has(oldID);
		if (streamExists) await this.destroy(oldID);
		if (!this.getSharedID(oldID)) {
			await _fs.FS.call(void 0, `${this.logPath}/modlog_${oldID}.txt`).rename(`${this.logPath}/modlog_${newID}.txt`);
		}
		if (streamExists) this.initialize(newID);
	}

	getActiveStreamIDs() {
		return [...this.streams.keys()];
	}

	/******************************************
	 * Methods for reading (searching) modlog *
	 ******************************************/
	 async runSearch(
		rooms, regexString, maxLines, onlyPunishments
	) {
		const useRipgrep = await checkRipgrepAvailability();
		let fileNameList = [];
		let checkAllRooms = false;
		for (const roomid of rooms) {
			if (roomid === 'all') {
				checkAllRooms = true;
				const fileList = await _fs.FS.call(void 0, this.logPath).readdir();
				for (const file of fileList) {
					if (file !== 'README.md' && file !== 'modlog_global.txt') fileNameList.push(file);
				}
			} else {
				fileNameList.push(`modlog_${roomid}.txt`);
			}
		}
		fileNameList = fileNameList.map(filename => `${this.logPath}/${filename}`);

		if (onlyPunishments) {
			regexString = `${onlyPunishments === 'global' ? GLOBAL_PUNISHMENTS_REGEX_STRING : PUNISHMENTS_REGEX_STRING}${regexString}`;
		}

		const results = new SortedLimitedLengthList(maxLines);
		if (useRipgrep) {
			if (checkAllRooms) fileNameList = [this.logPath];
			await this.runRipgrepSearch(fileNameList, regexString, results, maxLines);
		} else {
			const searchStringRegex = new RegExp(regexString, 'i');
			for (const fileName of fileNameList) {
				await this.readRoomModlog(fileName, results, searchStringRegex);
			}
		}
		return results.getListClone().filter(Boolean);
	}

	async runRipgrepSearch(paths, regexString, results, lines) {
		let output;
		try {
			const options = [
				'-i',
				'-m', '' + lines,
				'--pre', 'tac',
				'-e', regexString,
				'--no-filename',
				'--no-line-number',
				...paths,
				'-g', '!modlog_global.txt', '-g', '!README.md',
			];
			output = await execFile('rg', options, {cwd: _path.normalize.call(void 0, `${__dirname}/../`)});
		} catch (error) {
			return results;
		}
		for (const fileName of output.stdout.split('\n').reverse()) {
			if (fileName) results.insert(fileName);
		}
		return results;
	}

	async getGlobalPunishments(user, days = 30) {
		const response = await exports.PM.query({
			rooms: ['global' ],
			regexString: this.escapeRegex(`[${toID(user)}]`),
			maxLines: days * 10,
			onlyPunishments: 'global',
		});
		return response.length;
	}

	generateRegex(search) {
		// Ensure the generated regex can never be greater than or equal to the value of
		// RegExpMacroAssembler::kMaxRegister in v8 (currently 1 << 16 - 1) given a
		// search with max length MAX_QUERY_LENGTH. Otherwise, the modlog
		// child process will crash when attempting to execute any RegExp
		// constructed with it (i.e. when not configured to use ripgrep).
		return `[^a-zA-Z0-9]?${[...search].join('[^a-zA-Z0-9]*')}([^a-zA-Z0-9]|\\z)`;
	}

	escapeRegex(search) {
		return search.replace(/[\\.+*?()|[\]{}^$]/g, '\\$&');
	}

	async search(
		roomid = 'global',
		search = {},
		maxLines = 20,
		onlyPunishments = false
	) {
		const rooms = (roomid === 'public' ?
			[...Rooms.rooms.values()]
				.filter(room => !room.settings.isPrivate && !room.settings.isPersonal)
				.map(room => room.roomid) :
			[roomid]);

		// Ensure regexString can never be greater than or equal to the value of
		// RegExpMacroAssembler::kMaxRegister in v8 (currently 1 << 16 - 1) given a
		// searchString with max length MAX_QUERY_LENGTH. Otherwise, the modlog
		// child process will crash when attempting to execute any RegExp
		// constructed with it (i.e. when not configured to use ripgrep).
		let regexString = '.*?';
		if (search.anyField) regexString += `${this.escapeRegex(search.anyField)}.*?`;
		if (search.action) regexString += `\\) .*?${this.escapeRegex(search.action)}.*?: .*?`;
		if (search.user) {
			const wildcard = search.user.isExact ? `` : `.*?`;
			regexString += `.*?\\[${wildcard}${this.escapeRegex(search.user.search)}${wildcard}\\].*?`;
		}
		if (search.ip) regexString += `${this.escapeRegex(`[${search.ip}`)}.*?\\].*?`;
		if (search.actionTaker) regexString += `${this.escapeRegex(`by ${search.actionTaker}`)}.*?`;
		if (search.note) {
			const regexGenerator = search.note.isExact ? this.generateRegex : this.escapeRegex;
			for (const noteSearch of search.note.searches) {
				regexString += `${regexGenerator(toID(noteSearch))}.*?`;
			}
		}

		const query = {
			rooms: rooms,
			regexString,
			maxLines: maxLines,
			onlyPunishments: onlyPunishments,
		};
		const response = await exports.PM.query(query);

		if (response.duration > LONG_QUERY_DURATION) {
			Monitor.log(`Long modlog query took ${response.duration} ms to complete: ${query}`);
		}
		return {results: response, duration: response.duration};
	}

	 async readRoomModlog(path, results, regex) {
		const fileStream = _fs.FS.call(void 0, path).createReadStream();
		for await (const line of fileStream.byLine()) {
			if (!regex || regex.test(line)) {
				results.insert(line);
			}
		}
		void fileStream.destroy();
		return results;
	}
} exports.Modlog = Modlog;

// if I don't do this TypeScript thinks that (ModlogResult | undefined)[] is a function
// and complains about an "nexpected newline between function name and paren"
// even though it's a type not a function...


 const PM = new _processmanager.QueryProcessManager(module, async data => {
	const {rooms, regexString, maxLines, onlyPunishments} = data;
	try {
		const results = await exports.modlog.runSearch(rooms, regexString, maxLines, onlyPunishments);
		return results.map((line, index) => _converter.parseModlog.call(void 0, line, results[index + 1]));
	} catch (err) {
		Monitor.crashlog(err, 'A modlog query', data);
		return [];
	}
}); exports.PM = PM;

if (!exports.PM.isParentProcess) {
	global.Config = require('./config-loader').Config;
	global.toID = require('../.sim-dist/dex').Dex.toID;

	global.Monitor = {
		crashlog(error, source = 'A modlog process', details = null) {
			const repr = JSON.stringify([error.name, error.message, source, details]);
			// @ts-ignore please be silent
			process.send(`THROW\n@!!@${repr}\n${error.stack}`);
		},
	};

	process.on('uncaughtException', err => {
		if (Config.crashguard) {
			Monitor.crashlog(err, 'A modlog child process');
		}
	});

	// eslint-disable-next-line no-eval
	_repl.Repl.start('modlog', cmd => eval(cmd));
} else {
	exports.PM.spawn(MAX_PROCESSES);
}

 const modlog = new Modlog(MODLOG_PATH); exports.modlog = modlog;
